/*
   LibIntelCounters -- Support for reading Intel processor Counter
   Copyright (C) 2010 Leonardo de Paula Rosa Piga

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 3 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General
   Public License along with this library; if not, write to the
   Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301 USA
*/
#include "libIntelCounters.h"

#ifdef CORE2

void disableInstructionCounter(struct InstCounterSettings* s) {
  unsigned lsWord, msWord, event;
  unsigned long long ret;

  if (s->counter < 0 || s->counter >= MAX_COUNTER)
    return;
  // Read counter
  syscall(__NR_rdmsr, PERF_COUNTER_CONTROL, &ret);
  lsWord = (unsigned) ret;
  msWord = (unsigned) (ret >> 32);
  
  event = INSTR_COUNTER | ENB_MASK;
  if (s->user) event = event | USERC_MASK;
  if (s->sys) event = event | OSC_MASK;

  // Disable
  if (s->counter == 0) {
    lsWord = lsWord & ~ENABLE_PMC0;
    syscall(__NR_wrmsr, PERF_EVT_SEL0, event, 0x0);
    syscall(__NR_wrmsr, PERF_COUNTER_CONTROL, lsWord, msWord);
  } else {
    lsWord = lsWord & ~ENABLE_PMC1;
    syscall(__NR_wrmsr, PERF_EVT_SEL1, event, 0x0);
    syscall(__NR_wrmsr, PERF_COUNTER_CONTROL, lsWord, msWord);
  }
}

void enableInstructionCounter(struct InstCounterSettings* s) {
  unsigned lsWord, msWord, event;
  unsigned long long ret;

  if (s->counter < 0 || s->counter >= MAX_COUNTER)
    return;

  // Read counter
  syscall(__NR_rdmsr, PERF_COUNTER_CONTROL, &ret);
  lsWord = (unsigned) ret;
  msWord = (unsigned) (ret >> 32);

  event = INSTR_COUNTER | ENB_MASK;
  if (s->user) event = event | USERC_MASK;
  if (s->sys) event = event | OSC_MASK;

  // Enable
  if (s->counter == 0) {
    lsWord = lsWord | ENABLE_PMC0;
    syscall(__NR_wrmsr, PERF_EVT_SEL0, event, 0x0);
    syscall(__NR_wrmsr, PERF_COUNTER_CONTROL, lsWord, msWord);
  } else {
    lsWord = lsWord | ENABLE_PMC1;
    syscall(__NR_wrmsr, PERF_EVT_SEL1, event, 0x0);
    syscall(__NR_wrmsr, PERF_COUNTER_CONTROL, lsWord, msWord);
  }
}

unsigned long long instructionCounter(struct InstCounterSettings* s) {
  unsigned long long ret;
  unsigned eax, ecx, edx;                                             

  // Check bounds
  if (s->counter < 0 || s->counter >= MAX_COUNTER)
    return 0;

  ecx = (unsigned) s->counter;
  asm volatile("movl %2, %%ecx\n\t" 
               "rdpmc\n\t"          
               "movl %%eax, %0 \n\t"
               "movl %%edx, %1 \n\t"
               : "=a" (eax), "=d" (edx)
               : "c" (ecx));           
  ret = edx & 0xff;                    
  ret = (ret << 32) | eax;
  return ret;
}

unsigned int resetInstructionCounter(struct InstCounterSettings* s) {
  volatile unsigned eax, ebx, ecx, edx;
  unsigned long long ret;
  // Check bounds
  if (s->counter < 0 || s->counter >= MAX_COUNTER)
    return 0;

  do {
    disableInstructionCounter(s);
    // Set to zero
    if (s->counter == 0)
      syscall(__NR_wrmsr, GENERAL_PERF_COUNTER0, 0x0,0x0);
    else
      syscall(__NR_wrmsr, GENERAL_PERF_COUNTER1, 0x0,0x0);

    // Clear overflow bit
    syscall(__NR_rdmsr, PERF_COUNTER_GLOBAL_STATUS, &ret);
    ret = (1 << s->counter);
    syscall(__NR_wrmsr, PERF_COUNTER_OVF, (unsigned) ret, 0x40000000);

    syscall(__NR_rdmsr, PERF_COUNTER_GLOBAL_STATUS, &ret);
    ret &= (1 << s->counter);
  } while (ret); 

  // Repeat until we flush the pipeline and are able to reset the counter
  do {
    // Used to flush the pipeline
    asm volatile ("cpuid" : "=a" (eax), "=b" (ebx), "=c" (ecx), "=d" (edx) : "0" (0));
    if (s->counter == 0)
      syscall(__NR_rdmsr, GENERAL_PERF_COUNTER0, &ret);
    else
      syscall(__NR_rdmsr, GENERAL_PERF_COUNTER1, &ret);

    ecx = (unsigned) ret;
  } while (ecx != 0);

  if (s->counter == 0)
    syscall(__NR_rdmsr, GENERAL_PERF_COUNTER0, &ret);
  else
    syscall(__NR_rdmsr, GENERAL_PERF_COUNTER1, &ret);

  ecx = (unsigned) ret;
  return ecx;
}

//======================================================

unsigned long long cpuCoreCyclesCounter() {
  unsigned long long ret;
  syscall(__NR_rdmsr, CPU_CLK_UNHALTED_CORE, &ret);
  return ret;
}


unsigned int resetCpuCoreCyclesCounter() {
  unsigned long long ret;

  do {
    disableCpuCoreCyclesCounter();
    // Set to zero
    syscall(__NR_wrmsr, CPU_CLK_UNHALTED_CORE, 0x0,0x0);
    // Clear overflow bit
    syscall(__NR_wrmsr, PERF_COUNTER_OVF, 0x0, 0x40000002);
    // Check overflow bit
    syscall(__NR_rdmsr, PERF_COUNTER_GLOBAL_STATUS, &ret);
    ret = ret >> 32;
    ret &= 0x2;
  } while (ret); 

  syscall(__NR_rdmsr, CPU_CLK_UNHALTED_CORE, &ret);

  return (unsigned) ret;
}


void disableCpuCoreCyclesCounter() {
  unsigned lsWord, msWord;
  unsigned long long ret;

  // Read Perf Counter
  syscall(__NR_rdmsr, PERF_COUNTER_CONTROL, &ret);
  lsWord = (unsigned) ret;
  msWord = (unsigned) (ret >> 32);
  // Disable Perf Counter Control Enable Bit
  msWord = msWord & ~ENABLE_CPU_CLK_UNHALTED_CORE;
  syscall(__NR_wrmsr, PERF_COUNTER_CONTROL, lsWord, msWord);

  // Read Fixed Counter Control
  syscall(__NR_rdmsr, FIXED_COUNTER_CONTROL, &ret);
  lsWord = (unsigned) ret;
  msWord = (unsigned) (ret >> 32);

  // Disabe Fixed Counter Control Enable Bit
  lsWord = lsWord & ~CYCLES_CONTROL_MASK;
  syscall(__NR_wrmsr, FIXED_COUNTER_CONTROL, lsWord, msWord);
}


void enableCpuCoreCyclesCounter(struct FixedCounterSettings* s) {
  unsigned lsWord, msWord;
  unsigned long long ret;

  // Read Perf Counter
  syscall(__NR_rdmsr, PERF_COUNTER_CONTROL, &ret);
  lsWord = (unsigned) ret;
  msWord = (unsigned) (ret >> 32);
  // Set Perf Counter Control Enable Bit
  msWord = msWord | ENABLE_CPU_CLK_UNHALTED_CORE;
  syscall(__NR_wrmsr, PERF_COUNTER_CONTROL, lsWord, msWord);
  

  // Read Fixed Counter Control
  syscall(__NR_rdmsr, FIXED_COUNTER_CONTROL, &ret);
  lsWord = (unsigned) ret;
  msWord = (unsigned) (ret >> 32);
  // Set Fixed Counter Control Enable Bit
  if (s->user) lsWord = lsWord | CYCLES_CONTROL_USER_COUNT;
  if (s->sys) lsWord = lsWord | CYCLES_CONTROL_SYS_COUNT;
  syscall(__NR_wrmsr, FIXED_COUNTER_CONTROL, lsWord, msWord);

}

#endif /* CORE2 */
